#ifndef DUNE_PYTHON_GRID_COMMOPTS_HH
#define DUNE_PYTHON_GRID_COMMOPTS_HH

namespace Dune
{
  namespace Python
  {
    namespace detail
    {
      enum CommOp { add=0, set=1 };
    }
  }
}

#endif // DUNE_PYTHON_GRID_COMMOPTS_HH
